import json

from mcp.server.fastmcp import FastMCP

from wizlib.parser import WizParser
from wizlib.command import WizHelpCommand

from busy.command import BusyCommand


class MCPCommand(BusyCommand):

    name = 'mcp'

    @BusyCommand.wrap
    def execute(self):
        mcp = FastMCP("Busy")

        @mcp.tool()
        def perform(command: str) -> str:
            """Perform a Busy command"""
            try:
                result = self.app.do(command)
                return json.dumps(result)
            except Exception as error:
                return json.dumps({"Error": error})
        mcp.run()



from functools import cached_property
import re
from types import SimpleNamespace

from jinja2.sandbox import SandboxedEnvironment

from busy.command import CollectionCommand, MultiCollectionCommand
from busy.error import BusyError
from busy.model.item import FORMATTABLE_FIELDS, Item

from wizlib.parser import WizParser

VALID_FIELDS = {'num'} | FORMATTABLE_FIELDS


class ViewCommand(MultiCollectionCommand):
    """View items from a queue. Default format is just the 'simple' field.
    """

    name = 'view'
    unique: bool = False

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        formats = parser.add_mutually_exclusive_group()
        formats.add_argument(
            '--fixed', '-f', help='comma-separated fields to output')
        formats.add_argument(
            '--format', '-t', help='Python str.format template for each item')
        formats.add_argument(
            '--jinja', '-j', help='Jinja2 template for each item')
        formats.add_argument(
            '--list', '-i', action='store_true',
            help='List view with index numbers')
        parser.add_argument(
            '--unique', '-u', action='store_true',
            help='Performs a mildly fuzzy deduplication of output items')

    @CollectionCommand.wrap
    def execute(self):
        # Check for tag interdependency warnings
        # self.check_and_set_validation_warnings()

        if self.provided('jinja'):
            with open(self.jinja) as jinjafile:
                jinja = jinjafile.read()
                template = SandboxedEnvironment().from_string(jinja)
                items = [SimpleNamespace(**vals(m, x))
                         for x, m in self.viewable_items]
                return template.render(items=items)
        else:
            if self.provided('fixed'):
                row = self.fixed_format()
            elif self.provided('format'):
                row = self.format
            elif self.provided('list') and self.list:
                row = '{num:>6}  {listable}'
            else:
                row = '{base}'
            return '\n'.join(format_row(row, r, i) for i,
                             r in self.viewable_items)

    def fixed_format(self):
        """Format string for fixed width output"""
        cols = range(len(self.fields))
        formats = []
        for col in cols:
            field = self.fields[col]
            width = 1
            alignment = '<'
            for index, item in self.viewable_items:
                if field == 'num':
                    newwidth = len(str(index + 1))
                    alignment = '>'
                else:
                    # formattable = item.formattable
                    # if field in formattable and \
                    #         isinstance(formattable[field], int):
                    #     alignment = '>'
                    newwidth = colwidth(field, item)
                width = max(width, newwidth)
            formats += [f"{{{field}:{alignment}{width}}}"]
        return ' '.join(formats)

    @property
    def viewable_items(self):
        """Filter for uniqueness if requested"""
        if self.unique:
            items = []
            fuzzmatches = set()
            for index, item in reversed(self.selection):
                fuzzkey = item.fuzzkey
                if fuzzkey not in fuzzmatches:
                    items.append((index, item))
                    fuzzmatches.add(item.fuzzkey)
            items.reverse()
            return items
        else:
            return self.selection

    @property
    def fields(self):
        """Which fields are included in the output, validated - currently only
        works for fixed format."""
        fields = self.fixed.split(',')
        unknown_fields = [f for f in fields if
                          f.split('.')[0] not in VALID_FIELDS]
        if any(unknown_fields):
            raise BusyError(f"Unknown field(s) {','.join(unknown_fields)}")
        if 'num' in self.fixed and not self.collection.sequence_number_ok:
            raise BusyError('Invalid field num for multi-state view')
        return fields


def colwidth(field: str, item: Item) -> int:
    """Return the width required for a column, given the descriptor of a field
    and an item"""
    return len(('{' + field + '}').format(**item.formattable))


def vals(item: Item, index: int) -> dict:
    return item.formattable | {'num': index + 1}


def format_row(template: str, item: Item, index: int) -> str:
    return template.format(**vals(item, index)).rstrip()



from datetime import datetime
from functools import cached_property
from importlib import import_module

from wizlib.command import WizCommand
from wizlib.parser import WizParser
from wizlib.ui import Choice, Chooser
from wizlib.command import CommandCancellation

from busy.model.collection import Collection
from busy.error import BusyError
from busy.model.configured import ConfiguredQueue


class BusyCommand(WizCommand):

    default = 'simple'

    # TODO: Move to wizlib
    @staticmethod
    def add_yes_arg(parser: WizParser):
        parser.add_argument('--yes', '-y', action='store_true', default=None)

    # TODO: Move to wizlib
    def confirm(self, verb, *other_actions):
        """Ensure that a command is confirmed by the user"""
        if self.provided('yes'):
            return self.yes
        else:
            def cancel():
                raise CommandCancellation('Cancelled')
            chooser = Chooser(f"{verb}?", 'OK', [
                Choice('OK', '\n', True),
                Choice('cancel', 'c', cancel)
            ])
            for action in other_actions:
                name = action.name if hasattr(action, 'name') else 'other'
                key = action.key if hasattr(action, 'key') else 'o'
                chooser.add_choice(name, key, action)
            choice = self.app.ui.get_option(chooser)
            if type(choice) is bool:
                self.yes = choice
            return choice


class QueueCommand(BusyCommand):
    """Base for commands that work on the default collection of one queue"""

    queue_name: str = 'tasks'
    collection_state: str = 'todo'
    filter: list = None
    default_filter = [1]
    named_filters: list = []
    is_writer = False

    @property
    def collection(self):
        """Return the collection object being queried, usually todo"""
        if not hasattr(self, '_collection'):
            self._collection = self.app.storage.get_collection(
                self.queue_name, self.collection_state)
        return self._collection

    @property
    def complete_filter(self):
        """Simple filters plus named filters"""
        return self.filter + self.named_filters

    @property
    def selection(self):
        """Pairs (index, item) selected by filter"""
        return self.collection.selection(
            *self.complete_filter)

    @property
    def selected_indices(self):
        """Indices selected by filter"""
        return [i for i, t in self.selection]

    @property
    def selected_items(self):
        """Items in the selection"""
        return [t for i, t in self.selection]

    @property
    def selected_items_list(self):
        """Simple text list of selected items"""
        return '\n'.join([i.listable for i in self.selected_items])

    def time_value(self, minutes):
        """Estimated value of effort, usable for billable hours, rounded to one
        decimal place"""
        if not hasattr(self, '_multiplier'):
            self._multiplier = self.app.config.get('busy-multiplier') or 1.0
        adjusted_minutes = int(self._multiplier * minutes)
        return f"{adjusted_minutes//60}h{adjusted_minutes % 60}m"

    def summarize(self, items: list = None):
        if items is None:
            items = self.selected_items
        if len(items) == 1:
            result = "1 item"
        elif len(items) > 1:
            result = str(len(items)) + " items"
        else:
            return "nothing"
        elapsed = sum(i.elapsed_minutes for i in items)
        if elapsed:
            result += f" ({self.time_value(elapsed)})"
        return result

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--queue', '-q', default='tasks', nargs='?',
                            dest='queue_name')
        parser.add_argument('filter', nargs="*")

    @cached_property
    def queue(self):
        """The object holding the Queue configuration including tag
        configuration"""
        return self.app.get_queue(self.queue_name)

    def validate_selection(self):
        for item in self.selected_items:
            self.queue.validate_tags(item.tags)

    @property
    def is_timed(self):
        """Test if this operation requires timing"""
        return self.queue_name == 'tasks' and \
            self.is_writer

    def handle_vals(self):
        """Apply default filter and stop the timer"""
        super().handle_vals()
        self.queue.validate_is_active()
        # self.validate_queue_active()
        if not self.provided('filter'):
            self.filter = self.default_filter
        if self.is_timed:
            self.stop_current_task_timer()

    # We use a crude system for measuring elapsed time, which involves setting
    # the start_time and elapased_minutes values on the first item in the tasks
    # queue

    def stop_current_task_timer(self):
        """Stop the timer on the current top task before performing an
        operation, if we're working with the tasks queue"""
        todos = self.app.storage.get_collection('tasks', 'todo')
        if len(todos) > 0:
            todos[0].stop_timer()
            todos.changed = True

    def start_current_task_timer(self):
        """Start the timer on the current top task after performing an
        operation, if we're working with the tasks queue"""
        todos = self.app.storage.get_collection('tasks', 'todo')
        if len(todos) > 0:
            todos[0].start_timer()
            todos.changed = True

    @BusyCommand.wrap
    def execute(self, method, *args, **kwargs):
        """Handle timing and save operation"""
        result = method(self, *args, **kwargs)
        if self.is_timed:
            self.start_current_task_timer()
        self.app.storage.save()
        return (result if result else None)

    def set_next_item_status(self):
        """Sets the status to the top item in the todo collection of this
        command's queue for convenient output"""
        todos = self.app.storage.get_collection(self.queue_name, 'todo')
        self.status = todos[0].simple if len(todos) else None

    def output_items(self, func, with_index=False):
        """Return some attribute of all the items in the collection"""
        if with_index:
            return '\n'.join(func(self.collection[i], i)
                             for i in self.selected_indices)
        else:
            return '\n'.join(func(i) for i in self.selected_items)


class CollectionCommand(QueueCommand):
    """Base for commands that work on a user-specified collection"""

    states = ['done', 'todo', 'plan']

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument(
            '--state', '-s', action='store', default='todo',
            dest='collection_state', choices=cls.states)


class MultiCollectionCommand(CollectionCommand):
    """Base for commands that work on an optional user-specified collection. If
    no collection is specified, then numeric filters don't work and the command
    must be read-only."""

    states = ['done', 'todo', 'plan', 'multi']
    default_filter = []

    @property
    def collection(self):
        """Return the collection object being queried, which could include
        multiple states"""
        if self.collection_state == 'multi':
            if not hasattr(self, '_collection'):
                self._collection = Collection()
                for state in self.states[:-1]:
                    collection = self.app.storage.get_collection(
                        self.queue_name, state)
                    if collection:
                        self._collection += [i for i in collection]
            return self._collection
        else:
            return super().collection


class NullCommand(QueueCommand):
    name = 'null'

    @QueueCommand.wrap
    def execute(self):
        pass


class IntegrationCommand:

    def execute_integration(self):
        module = import_module('busy_' + self.integration)
        integration = module.Main()
        func = getattr(integration, self.name)
        result = func(self)
        return result



# We're slowly deprecating this command in favour of the view command

from wizlib.parser import WizParser

from busy.command import CollectionCommand
from busy.util.date_util import absolute_date


class ListCommand(CollectionCommand):
    """Show the markup with selection numbers, default to all. Includes summary
    output."""

    name = 'list'
    default_filter = ['1-']
    FORMATS = {
        'markup': "{!s}",
        'plan_date': "{:%Y-%m-%d}",
        'done_date': "{:%Y-%m-%d}"
    }

    def handle_vals(self):
        super().handle_vals()
        self.named_filters = []
        if self.provided('done_min'):
            def minfunc(i): return i.done_date >= absolute_date(self.done_min)
            self.named_filters.append(minfunc)
        if self.provided('done_max'):
            def maxfunc(i): return i.done_date <= absolute_date(self.done_max)
            self.named_filters.append(maxfunc)

    @CollectionCommand.wrap
    def execute(self):
        def format(item, index):
            result = f"{(index+1):>6}"
            for colname in self.collection.schema:
                format = self.FORMATS[colname]
                if (colname == 'markup'):
                    value = item.listable
                else:
                    value = getattr(item, colname)
                result += f"  {format.format(value)}"
            return result

        # Check for tag interdependency warnings
        # self.check_and_set_validation_warnings(
        #     lambda: setattr(self, 'status', self.summarize()))
        self.status = self.summarize() if self.selected_items else ''

        return self.output_items(format, with_index=True)



from busy.command import CollectionCommand
from busy.model.collection import Collection


class TagsCommand(CollectionCommand):

    name = 'tags'

    @CollectionCommand.wrap
    def execute(self):
        # Check for tag interdependency warnings
        # self.check_and_set_validation_warnings()

        tags = set()
        for item in self.collection:
            tags |= item.tags
        return '\n'.join(sorted(tags))



import yaml
from functools import cached_property

from busy.command import MultiCollectionCommand
from busy.model.item import Item


class DumpCommand(MultiCollectionCommand):
    """Dump items as YAML with same filtering as view command"""

    name = 'dump'

    @MultiCollectionCommand.wrap
    def execute(self):
        # Check for tag interdependency warnings
        # self.check_and_set_validation_warnings()

        items_data = []
        for index, item in self.viewable_items:
            item_dict = self._item_to_dict(item)
            if item_dict:  # Only add if has content
                items_data.append(item_dict)

        result = {self.queue_name: items_data}
        return yaml.dump(result, default_flow_style=False, sort_keys=False)

    @property
    def viewable_items(self):
        return self.selection

    def _item_to_dict(self, item):
        """Convert an Item to a dictionary with only non-empty values"""
        result = {}

        # Always include base and state
        result['base'] = item.base
        result['state'] = item.state

        # Tags as list (only if not empty)
        if item.tags:
            result['tags'] = sorted(list(item.tags))

        # Data section (preserve original types)
        # data_dict = {}
        # for mark in item._marked("%"):
        #     key = mark[0]
        #     value = mark[1:]
        #     # Try to preserve type - if it's numeric, keep as number
        #     try:
        #         if '.' in value:
        #             data_dict[key] = float(value)
        #         else:
        #             data_dict[key] = int(value)
        #     except ValueError:
        #         data_dict[key] = value  # Keep as string
        # if data_dict:
        #     result['data'] = data_dict
        if item.vals:
            result['data'] = item.vals

        # Timing section (only if has timing data)
        timing_dict = {}
        if item.elapsed_minutes > 0:
            timing_dict['elapsed'] = item.elapsed_minutes
        if item.start_time:
            timing_dict['start'] = None  # ISO format when timing is active

        if timing_dict:
            result['timing'] = timing_dict

        # URL (only if not empty)
        if item.url:
            result['url'] = item.url

        # Repeat (only if not empty)
        if item.repeat:
            result['repeat'] = item.repeat

        # Dates section (only if has any dates)
        dates_dict = {}
        if item.done_date:
            dates_dict['done'] = item.done_date.strftime('%Y-%m-%d')
        if item.plan_date:
            dates_dict['plan'] = item.plan_date.strftime('%Y-%m-%d')

        if dates_dict:
            result['dates'] = dates_dict

        return result



from busy.command import CollectionCommand


class DescribeCommand(CollectionCommand):
    """Show the full markup"""

    name = 'describe'

    @CollectionCommand.wrap
    def execute(self):
        # Check for tag interdependency warnings
        # self.check_and_set_validation_warnings()

        return self.output_items(lambda i: i.markup)



import re
from dataclasses import KW_ONLY, dataclass, field
from datetime import date, datetime
from typing import Optional


from busy.util.date_util import absolute_date
from busy.util.date_util import relative_date
from busy.util import date_util


class ItemStateError(Exception):
    pass


# Some handy constants

START_TIME_FORMAT = '%Y%m%d%H%M'
DATE_FORMAT = '%Y-%m-%d'
VALID_STATES = {'todo', 'plan', 'done'}

# Helpers for parsing - older versions of Busy supported an elaborate
# "followon" mechanism

FOLLOW_SPLIT = re.compile(r'\s*\-*\>\s*')
LEGACY_REPEAT = re.compile(r'^\s*repeat(?:\s+[io]n)?\s+(.+)\s*$', re.I)


# Dumb object to simplify format string use

class Obj:
    def __getattr__(self, attr):
        return ''


# Function to simplify joining strings

def _joined(*substrings, sep: str = ' '):
    return sep.join([s for s in substrings if s])


# Decorator to restrict operations to certain states

def restricted(*allowed_states):
    def wrapper(method):
        def replacement(self, *args, **kwargs):
            if self.state in allowed_states:
                return method(self, *args, **kwargs)
            else:
                raise ItemStateError(
                    f"Method not allowed in state '{self.state}'")
        return replacement
    return wrapper


@dataclass
class Item:
    """A single entry in a queue"""

    _: KW_ONLY

    # A unique identifier for the item
    id: str = None

    # The basic text of the item
    base: str

    # The state, which must be valid
    state: str = 'todo'

    # Unique tags, indicated by `#` in markup
    tags: set[str] = field(default_factory=set)

    # Data values, indicated by `%` and a single letter in markup
    vals: dict[str, str] = field(default_factory=dict)

    # A string describing if and when to repeat the item
    repeat: str = ''

    # A single URL, indicated by `@` in markup
    url: str = ''

    # Number of clock minutes spent on the task according to our crude timing
    # system
    elapsed_minutes: int = 0

    # Clock time that this task started, only for top item in `tasks` queue,
    # indicated by `!s` in markup
    start_time: Optional[datetime] = None

    # Date the item can be activated, only if item in `plan` state
    plan_date: Optional[date] = None

    # Date the item was done, only if item in `done` state
    done_date: Optional[date] = None

    @classmethod
    def from_markup(cls, markup: str, state: str = 'todo',
                    plan_date: date | str | None = None,
                    done_date: date | str | None = None):
        split = FOLLOW_SPLIT.split(markup, maxsplit=1)
        body = split[0] if split else ""

        if len(split) > 1:
            next = split[1]
            match = LEGACY_REPEAT.match(next)
            if match:
                repeat = match.group(1)
            else:
                repeat = next
        else:
            repeat = ""

        words = body.split()
        base = " ".join([w for w in words if w[0] not in '#@!%?'])
        def marked(mark): return [w[1:] for w in words if w.startswith(mark)]
        id = mid[0] if (mid := marked('?')) else ''
        tags = {m.lower() for m in marked('#')}
        url = murl[0] if (murl := marked('@')) else ''
        vals = {m[0]: m[1:] for m in marked('%')}
        timing_data = {m[0]: m[1:] for m in marked("!")}
        elapsed_minutes = int(timing_data['e']) if 'e' in timing_data else 0
        start_time = datetime.strptime(timing_data['s'], START_TIME_FORMAT) \
            if 's' in timing_data else None
        return cls(
            id=id,
            state=state,
            base=base,
            repeat=repeat,
            tags=tags,
            url=url,
            vals=vals,
            elapsed_minutes=elapsed_minutes,
            start_time=start_time,
            plan_date=absolute_date(plan_date),
            done_date=absolute_date(done_date)
        )

    # --- Limit state to valid options ---

    def __post_init__(self):
        if self.state not in VALID_STATES:
            raise ValueError(f"Item state must be one of {VALID_STATES}")

    # ---- Helpers to reconstruct markup ----

    @property
    def _mid(self):
        """The ID is marked with a questionmark, even though it's always
        first"""
        return f"?{self.id}" if self.id else ''

    @property
    def _mtags(self):
        return sorted({f"#{t}" for t in self.tags})

    @property
    def _murl(self):
        return f"@{self.url}" if self.url else ''

    @property
    def _mvals(self):
        return {f"%{k}{v}" for k, v in self.vals.items()}

    @property
    def _melapsed(self):
        return f"!e{self.elapsed_minutes}" if self.elapsed_minutes else ''

    @property
    def _mstart(self):
        return f"!s{self.start_time.strftime(START_TIME_FORMAT)}" \
            if self.start_time else ''

    # ---- Output forms, including markup subcomponents ----

    @property
    def noparens(self):
        """Base text without parenthetical expressions"""
        result = ''
        depth = 0
        for char in self.base:
            if char == '(':
                depth += 1
            elif char == ')' and depth > 0:
                depth -= 1
            elif depth == 0:
                result += char
        return result.strip()

    @property
    def simple(self):
        """Base text plus marked tags"""
        return _joined(self.base, *self._mtags)

    @property
    def listable(self):
        """Simple plus repeat"""
        return _joined(self.simple, self.repeat, sep=' > ')

    @property
    def notiming(self):
        """Base text with tags, vals, and URL"""
        return _joined(self.simple, self._murl, *self._mvals)

    @property
    def body(self):
        """Everything except the repeat"""
        return _joined(self.notiming, self._melapsed, self._mstart)

    @property
    def checkbox(self):
        """GitLab-style Markdown checkbox"""
        checked = 'x' if self.state == 'done' else ' '
        return f"- [{checked}]"

    @property
    def fuzzkey(self):
        """Key for fuzzy matching and deduplication"""
        return self.noparens.lower()

    @property
    def markup(self):
        main = _joined(self._mid, self.body)
        return _joined(main, self.repeat, sep=' > ')

    def __str__(self):
        """Represent the item as its simple form"""
        return self.simple

    # ---- State operations ----

    @restricted('todo')
    def done(self, done_date: date, plan_date: date = None):
        """Updates the item to done and returns a copy as a plan for the
        plan_date if provided"""
        plan = None
        if plan_date:
            plan_props = {
                'state': 'plan',
                'plan_date': plan_date,
                'elapsed_minutes': None,
                'start_time': None
            }
            plan = Item(**(vars(self) | plan_props))
        self.state = 'done'
        self.done_date = done_date
        self.repeat = ''
        return plan

    @restricted('done')
    def undone(self):
        self.state = 'todo'

    @restricted('todo')
    def plan(self, plan_date: date):
        self.state = 'plan'
        self.plan_date = plan_date

    @restricted('plan')
    def unplan(self):
        self.state = 'todo'

    # ---- Timer operations ----

    @restricted('todo')
    def start_timer(self):
        """Starts timing when activity begins on a task - is really only ever
        called on the top todo task"""
        if self.base and not self.start_time:
            self.start_time = date_util.now()

    @restricted('todo')
    def stop_timer(self):
        """Stop the timer and update elapsed minutes, typically called before
        change operations"""
        if self.start_time:
            if self.base:
                now = date_util.now()
                prev_elapsed = self.elapsed_minutes
                new = (now - self.start_time).seconds // 60
                self.elapsed_minutes = new + prev_elapsed
            self.start_time = None

    # ---- Named filters for Selector ----

    def filter_val(self, key_val):
        val = self.vals[k] if (k := key_val[0]) in self.vals else None
        return ((val is not None) and (val == key_val[1:]))

    def filter_donemin(self, min_date):
        if self.done_date:
            return self.done_date >= absolute_date(min_date)

    def filter_donemax(self, max_date):
        if self.done_date:
            return self.done_date <= absolute_date(max_date)

    def filter_planmin(self, min_date):
        if self.plan_date:
            return self.plan_date >= absolute_date(min_date)

    def filter_planmax(self, max_date):
        if self.plan_date:
            return self.plan_date <= absolute_date(max_date)

    # ---- Properties for reading in formatted output

    @property
    def _otags(self):
        """Object-like format for tags"""
        result = Obj()
        for tag in self.tags:
            setattr(result, tag, tag)
        return result

    @property
    def _ovals(self) -> Obj:
        """Object-like format for data vals"""
        result = Obj()
        for key in self.vals:
            setattr(result, key, self.vals[key])
        return result

    @property
    def _sdonedate(self) -> str:
        if self.state == 'done' and self.done_date:
            return self.done_date.strftime(DATE_FORMAT)

    @property
    def _splandate(self) -> str:
        if self.state == 'plan' and self.plan_date:
            return self.plan_date.strftime(DATE_FORMAT)

    @property
    def formattable(self):
        return {
            'tag': self._otags,
            'tags': ' '.join(self.tags),
            'val': self._ovals,
            'donedate': self._sdonedate,
            'plandate': self._splandate,
            'elapsed': str(self.elapsed_minutes)
        } | {k: getattr(self, k) for k in {'body', 'base', 'noparens',
                                           'simple', 'listable', 'id',
                                           'notiming', 'url', 'repeat',
                                           'markup', 'checkbox'}}


FORMATTABLE_FIELDS = {'id', 'body', 'base', 'noparens', 'simple', 'listable',
                      'notiming', 'url', 'repeat', 'markup', 'tags', 'tag',
                      'val', 'donedate', 'plandate', 'elapsed', 'checkbox'}



import os
import fcntl
import time
from pathlib import Path


# Utility to manage unique item identifiers that can be used to enable
# multi-device use and data merging.
#
# The scope of the identifiers includes:
#
# - **domain** included in configuration (and not actually implemented) with
#   the idea that a domain is a single user in a single context, and that
#   merges can only happen between addresses on the same domain. So the domain
#   is the namespace for identifiers.
# - **address* really just a device identifier, though could be used for
#   multiple addresses on one device (for example, to hold personal and work
#   data sets). Also stored in the configuration. There are up to 32 addresses
#   within a domain. Addresses are shown in all identifiers as the first
#   character of the identifier. For example, if the identifier is 'B64JW9'
#   then the address is 'B'.
# - **counter** a 5-character base-32 counter than is incremented each time a
#   new identifier is needed. The latest counter value is stored alongside the
#   data files.

BASE32_ALPHABET = '0123456789ABCDEFGHJKLMNPQRSTUVWX'
COUNTER_WIDTH = 5
COUNTER_FILENAME = 'counter'
COUNTER_START = '00000'


class Identifier:
    def __init__(self, address: str, directory: str, width=COUNTER_WIDTH):
        self.width = width
        if address not in BASE32_ALPHABET:
            raise ValueError(
                f"Invalid device address {address}"
            )
        self.address = address
        self.file = Path(directory) / COUNTER_FILENAME
        if not self.file.exists():
            self.file.parent.mkdir(parents=True, exist_ok=True)
            with open(self.file, 'w') as f:
                f.write(COUNTER_START)

    def _encode(self, number: int) -> str:
        """Convert integer to fixed-width base32 string"""
        if number >= (32 ** self.width):
            raise ValueError(
                f"Value {number} exceeds {self.width} characters")
        result = ''
        for _ in range(self.width):
            result = BASE32_ALPHABET[number % 32] + result
            number //= 32
        return result

    def _decode(self, base32: str) -> int:
        """Convert base32 string to integer"""
        if len(base32) != self.width:
            raise ValueError(
                f"Identifiers must have {self.width} characters"
            )
        result = 0
        for char in base32:
            if char not in BASE32_ALPHABET:
                raise ValueError(f"Invalid base32 character: {char}")
            result = result * 32 + BASE32_ALPHABET.index(char)
        return result

    def increment(self):
        """Read current identifier, increment counter, and return new
        identifier"""
        with open(self.file, 'r+') as f:
            fcntl.flock(f.fileno(), fcntl.LOCK_EX)
            try:
                f.seek(0)
                prev = f.read().strip()
                if not prev:
                    raise RuntimeError(f'Data missing from {self.file}')
                next = self._encode(self._decode(prev) + 1)
                f.seek(0)
                f.write(next)
                f.truncate()
                f.flush()
                os.fsync(f.fileno())  # Force write to disk
                return f"{self.address}{next}"
            finally:
                pass



import re

from jinja2.sandbox import SandboxedEnvironment

from busy.command import CollectionCommand, MultiCollectionCommand
from busy.error import BusyError
from busy.model.item import Item

from wizlib.parser import WizParser


class ReportCommand(MultiCollectionCommand):
    """Output items using a defined string format."""

    name = 'report'

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--line', '-l')

    @CollectionCommand.wrap
    def execute(self):
        items = self.selection
        result = [self.formatted(x, m) for x, m in items]
        return '\n'.join(result)

    def formatted(self, index, item):
        # return self.line.format(**vars(item))
        template = SandboxedEnvironment().from_string(self.line)
        return template.render(item.formattable)



from __future__ import annotations
from dataclasses import dataclass, field
from functools import cached_property

from busy.error import BusyError


ACTIVE_LABEL = 'active'

TAGS_LABEL = 'tags'


class Configured:

    active: bool = True

    def set_active(self, config):
        """Set the active attribute based on the config. If the config itself
        is boolean, that's the active value, otherwise check the active
        attribute."""
        if isinstance(config, bool):
            self.active = config
        elif isinstance(config, dict) and ACTIVE_LABEL in config:
            self.active = config[ACTIVE_LABEL]
        else:
            self.active = True

    # def set_tags(self, config, queue):
    #     """Add a set of tags, and index them in the queue"""
    #     if isinstance(config, dict) and TAGS_LABEL in config:
    #         tags = {n:ConfiguredTag.from_config(n, c, tag) for
    # n,c in config[TAGS_LABEL]}


@dataclass
class ConfiguredQueue(Configured):

    name: str

    # The tags dictionary contains reference to all the configured tags
    configured_tags: dict[str, ConfiguredTag] = field(default_factory=dict)

    # Similar structure to tag.from_config - possible optimization opportunity?

    @classmethod
    def from_config(cls, name: str, config: dict | bool = None):
        queue = cls(name)
        queue.set_active(config)
        if isinstance(config, dict) and TAGS_LABEL in config:
            for tag_name in (tags_config := config[TAGS_LABEL]):
                ConfiguredTag.parse_config(
                    tag_name, tags_config[tag_name],
                    parent_or_queue=queue)
        return queue

    def add_tag(self, configured_tag):
        name = configured_tag.name
        if name in self.configured_tags:
            raise BusyError(
                f'Duplicate configured tag {name} in queue {self.name}')
        self.configured_tags[name] = configured_tag

    def get_tag(self, name):
        """Return the tag object for this queue if it exists - note that
        queue.tags is only the root level of the hierarchy"""
        return next(
            (x for t in self.configured_tags.values()
             if (x := t.get_tag(name))),
            None)

    def validate_is_active(self):
        if not self.active:
            raise BusyError(f"Queue {self.name} is inactive")

    def validate_tags(self, tags: set[str]):
        """Confirm that the tags passed in fit the hierarchy defined in the
        config (if any)"""
        if not isinstance(tags, set):
            tags = set(tags)
        for tag in tags:
            if configured_tag := self.get_tag(tag):
                if configured_tag.active is False:
                    raise BusyError(
                        f"Tag {tag} is inactive for queue {self.name}")
                if options := configured_tag.active_child_tags:
                    if len(tags & options) != 1:
                        raise BusyError(
                            f"Queue '{self.name}' requires items " +
                            f"with tag '{tag}' to have exactly one of " +
                            ' '.join(f"'{t}'" for t in sorted(options)))
                if (parent := configured_tag.parent):
                    if not ((name := parent.name) in tags):
                        raise BusyError(
                            f"Queue '{self.name}' requires items "
                            f"with tag '{tag}' to have parent tag '{name}'")

    # def validate_tags_active(self, tags):
    #     """Validate that all tags in the item are
    # active according to config"""
    #     for tag in tags:
    #         self._validate_active(self.tags_config, tag, "Tag")

    # def tag_violations_exist(self, tags:set[str]):
    #     """Given a set of tags, return true if there are hierarchy violations
    #     for this queue"""
    #     for tag in tags:
    #         if tag in self.parent2children:
    #             # Parent tag must have exactly one child
    #             children_present = tags.intersection(
    #                 self.parent2children[tag])
    #             if len(children_present) != 1:
    #                 return True
    #         if tag in self.child2parent:
    #             # Child tag must have its parent
    #             if self.child2parent[tag] not in tags:
    #                 return True
    #     return False

    # @cached_property
    # def parent2children(self):
    #     """A dict of tags, where values are the set of
    # children of that tag"""

    # def tag_is_actice(self, tag: str):


@dataclass
class ConfiguredTag(Configured):

    queue: ConfiguredQueue

    name: str

    parent: ConfiguredTag | None = field(default=None, kw_only=True)

    children: dict[str, ConfiguredTag] = field(
        default_factory=dict, kw_only=True)

    @classmethod
    def parse_config(cls, name: str, config: dict | bool = None, *,
                     parent_or_queue: ConfiguredTag | ConfiguredQueue):
        if isinstance(parent_or_queue, ConfiguredQueue):
            tag = cls(parent_or_queue, name)
        elif isinstance(parent_or_queue, ConfiguredTag):
            tag = cls(parent_or_queue.queue, name, parent=parent_or_queue)
        tag.set_active(config)
        if isinstance(config, dict) and TAGS_LABEL in config:
            for tag_name in (tags_config := config[TAGS_LABEL]):
                cls.parse_config(
                    tag_name, tags_config[tag_name],
                    parent_or_queue=tag)
        return tag

    def __post_init__(self):
        self.queue.add_tag(self)
        if self.parent:
            self.parent.add_child(self)

    def add_child(self, configured_tag):
        self.children[configured_tag.name] = configured_tag

    def get_tag(self, name):
        """Return self or descendent tag by name"""
        if name == self.name:
            return self
        else:
            return next(
                (x for t in self.children
                 if (x := self.children[t].get_tag(name))),
                None)

    @cached_property
    def active_child_tags(self) -> set:
        return set(t.name for t in self.children.values() if t.active)



from functools import cached_property
from pathlib import Path

import sys
from wizlib.app import WizApp
from wizlib.ui_handler import UIHandler
from wizlib.config_handler import ConfigHandler
from wizlib.command import WizHelpCommand

from busy.command import BusyCommand
from busy.storage.file_storage import FileStorage
from busy.util.identifier import Identifier
from busy.model.configured import ConfiguredQueue


class BusyApp(WizApp):

    base = BusyCommand
    name = 'busy'
    handlers = [UIHandler, ConfigHandler]

    @cached_property
    def directory(self):
        return self.config.get('busy-storage-directory') \
            or Path.home() / '.busy'

    @cached_property
    def identifier(self):
        identifier_address = self.config.get('busy-identifier-address') \
            or '0'
        return Identifier(identifier_address, self.directory)

    @cached_property
    def storage(self):
        return FileStorage(self.directory, self.identifier)

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    @cached_property
    def queues(self):
        if (queues_config := self.config.get('busy-queues')):
            return {n: ConfiguredQueue.from_config(n, c) for
                    n, c in queues_config.items()}
        else:
            return {}

    def get_queue(self, name):
        if name in self.queues:
            return self.queues[name]
        elif '_' in self.queues:
            return self.queues['_']
        else:
            return ConfiguredQueue(name)

    # def run(self, **vals):
    #     super().run(**vals)

    # For MCP server. Probably belongs in WizLib.
    def do(self, command_string) -> dict:
        args = command_string.split(' ')
        ns = self.parser.parse_args(args)
        vals = vars(ns)
        if 'help' in vals:
            ccls = WizHelpCommand
        else:
            c = 'command'
            cname = (vals.pop(c) if c in vals else None) or self.base.default
            ccls = self.base.family_member('name', cname)
            if not ccls:
                raise Exception(f"Unknown command {cname}")
        command = ccls(self, **vals)
        result = command.execute()
        return {'result': result, 'status': command.status}



from wizlib.parser import WizParser
from wizlib.ui.shell_ui import Emphasis

from busy.command import QueueCommand
from busy.model.item import Item
from busy.util import date_util


class AddCommand(QueueCommand):

    name = 'add'
    is_writer = True
    pop: bool = False
    when: str = ""

    @classmethod
    def add_args(cls, parser: WizParser):
        # Special case, no filter argument
        parser.add_argument('--queue', '-q', default='tasks', nargs='?',
                            dest='queue_name')
        parser.add_argument('--pop', '-p', action='store_true', default=None)
        parser.add_argument('--when', '-w')
        parser.add_argument('markup', default="", nargs='?')

    def handle_vals(self):
        super().handle_vals()
        if not self.provided('markup'):
            self.markup = self.app.ui.get_text('Item: ')

    def check_when(self):
        plan_date = date_util.relative_date(self.when) if self.when else None
        if plan_date:
            return plan_date
        elif self.when:
            self.app.ui.send(f"Invalid time '{self.when}'", Emphasis.ERROR)
        return None

    @QueueCommand.wrap
    def execute(self):
        if self.markup:
            item = Item.from_markup(self.markup)
            self.queue.validate_tags(item.tags)
            # self.queue.validate_tags_active(item.tags)
            # self.validate_tags_active(item)
            # self.validate_tag_interdependencies([item])
            if self.provided('when'):
                plan_date = self.check_when()
                if plan_date:
                    plans = self.app.storage.get_collection(
                        self.queue_name, 'plan')
                    item.plan(plan_date)
                    plans.append(item)
                    self.set_next_item_status()
                    return
            if self.pop:
                self.collection.insert(0, item)
            else:
                self.collection.append(item)
        self.set_next_item_status()



# An enhanced list of ITems

from collections import UserList
from collections.abc import Iterable
from csv import DictReader, DictWriter
from io import IOBase

from busy.model.item import Item
from wizlib.class_family import ClassFamily
from busy.util.identifier import Identifier
from busy.util.selector import Selector


class CollectionError(Exception):
    pass


class Collection(UserList, ClassFamily):

    # Attributes in order for storage and editing
    schema = ['markup']

    # Do we allow numeric sequence number filters?
    sequence_number_ok = False

    def __init__(self, value=None):
        if value is None:
            super().__init__()
        else:
            self.validate(value)
            super().__init__(value)
        self.changed = False

    def validate_item(self, item):
        if not self.state_is_valid(item):
            raise CollectionError(
                f"Incorrect state {item.state} for {self.state} collection")
        if '|' in item.markup:
            raise CollectionError(f"Pipe symbol not permitted in item " +
                                  f"markup '{item.markup}'")

    def state_is_valid(self, item):
        return True

    def validate(self, value):
        for item in value if isinstance(value, Iterable) else [value]:
            self.validate_item(item)

    # Override list methods to make sure we've got Items of the right state.
    # There might be some methods missing here. Also set changed.

    def __setattr__(self, name, value):
        if name == 'data':
            self.validate(value)
        super().__setattr__(name, value)
        if name == 'data':
            self.changed = True

    def __setitem__(self, key, value):
        self.validate(value)
        self.changed = True
        return super().__setitem__(key, value)

    def __iadd__(self, value):
        self.validate(value)
        self.changed = True
        return super().__iadd__(value)

    def __delitem__(self, *args):
        self.changed = True
        return super().__delitem__(*args)

    def append(self, value):
        self.validate(value)
        self.changed = True
        return super().append(value)

    def extend(self, value):
        self.validate(value)
        self.changed = True
        return super().extend(value)

    def insert(self, index, value):
        self.validate(value)
        self.changed = True
        return super().insert(index, value)

    def __add__(self, value):
        self.validate(value)
        self.changed = True
        return super().__add__(value)

    def pop(self, *args):
        self.changed = True
        return super().pop(*args)

    # Methods related to selection and maniuplation by "filter"
    # Numbers in "filter" start with 1.

    def selection(self, *filter):
        """
        Take 'filter' (such as tags, ranges of numbers, etc) and return tuples
        of the indices and items selected. In other parts of the class and
        subclasses, operations need to know what items to act on. A collection
        that might contain multiple states doesn't allow sequence number
        filters.
        """
        selector = Selector(filter, sequence_number_ok=self.sequence_number_ok)
        return [(i, self.data[i]) for i in selector.indices(self.data)]

    def items(self, indices: list[int] = None):
        """Return the items for a set of indices"""
        if indices is None:
            return self.data
        else:
            return [self[i] for i in indices]

    def replace(self, indices: list, newvalues: list):
        """Replace existing items at the select provided.

        Also inserts if the indices run out. Does not create items; expects
        them to already exist. Does not return anything since the calling code
        already has the new items.
        """
        _indices = sorted(indices)
        _newvalues = newvalues.copy()
        max = (_indices[-1] + 1) if _indices else 0
        while _newvalues and _indices:
            self[_indices.pop(0)] = _newvalues.pop(0)
        while _indices:
            del self[_indices.pop()]
        self[max:max] = _newvalues

    def delete(self, indices: list[int]) -> list:
        """Remove the items at select and return them"""
        killlist, keeplist = self.split(indices)
        self.data = keeplist
        self.changed = True
        return killlist

    def split(self, indices: list[int]) -> tuple[list, list]:
        """Return a tuple of items at indices and items not so"""
        inlist = [t for i, t in enumerate(self) if i in indices]
        outlist = [t for i, t in enumerate(self) if i not in indices]
        return (inlist, outlist)

    def write_items(
            self, fileish: IOBase, indices: list = None,
            identifier: Identifier = None):
        """
        Write all or a set of items to a file-like object
        """
        writer = DictWriter(fileish, self.schema, delimiter="|")
        for item in self.items(indices):
            if identifier and not item.id:
                item.id = identifier.increment()
            values = dict([(f, getattr(item, f)) for f in self.schema])
            writer.writerow(values)


class SingleStateCollection(Collection):

    # With only one state, an index-type filter (number or range) is OK.
    sequence_number_ok = True

    def state_is_valid(self, item):
        """Just the state check"""
        return item.state == self.state

    def read_items(
            self, fileish: IOBase, indices: list = None,
            identifier: Identifier = None,
            from_storage: bool = False):
        """
        Read all or a set of items from a file-like object. Used for loading
        from storage and editing operations.
        """
        indices = indices if indices else []
        reader = DictReader(fileish, self.schema, delimiter="|")
        items = []
        changed = False
        for row in reader:
            item = Item.from_markup(state=self.state, **row)
            if identifier and not item.id:
                item.id = identifier.increment()
                changed = True
            items.append(item)
        self.replace(indices, items)
        if from_storage and not changed:
            self.changed = False
        return items
